﻿//-----------------------------------------------------------------------------
// Copyright (c) 2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Dhpoware
{
    /// <summary>
    /// Custom vertex structure used for billboarding.
    /// </summary>
    public struct BillboardVertex : IVertexType
    {
        public static readonly VertexDeclaration VertexDeclaration = new VertexDeclaration
        (
            new VertexElement(0, VertexElementFormat.Vector3, VertexElementUsage.Position, 0),
            new VertexElement(sizeof(float) * 3, VertexElementFormat.Vector4, VertexElementUsage.TextureCoordinate, 0)
        );

        public Vector3 Position;
        public Vector4 TexCoord;

        public BillboardVertex(Vector3 position, Vector2 texCoord, Vector2 offset)
        {
            Position = position;
            
            // Coordinates for the texture map.
            TexCoord.X = texCoord.X;
            TexCoord.Y = texCoord.Y;

            // The 2D offset vector.
            TexCoord.Z = offset.X;
            TexCoord.W = offset.Y;

        }

        VertexDeclaration IVertexType.VertexDeclaration
        {
            get { return VertexDeclaration; }
        }
    }

    /// <summary>
    /// A class to manage the resources required to implement billboarding
    /// using shaders. Each billboard consists of 4 vertices and 6 indices.
    /// The billboard's quad is rendered as an indexed triangle list consisting
    /// of 2 triangles. All of the vertices for the billboard's triangles are
    /// positioned at the center of the billboard - as specified in the
    /// Billboard class' constructor. The Billboard's texture coordinate
    /// vertex attribute is a 4 component floating point vector. The first 2
    /// components is the 2D texture coordinates of the color map. The other
    /// 2 components are the x and y offsets from the center of the billboard
    /// that the vertex shader will use to transform each vertex to its
    /// correct position.
    /// </summary>
    public class Billboard
    {
        public VertexBuffer vertexBuffer;
        private IndexBuffer indexBuffer;

        public Billboard(GraphicsDevice graphicsDevice, Vector3 position, float width, float height)
        {
            Create(graphicsDevice, position, width, height);
        }

        public void Create(GraphicsDevice graphicsDevice, Vector3 position, float width, float height)
        {
            // The billboard quad is made of 2 triangles and is centered about
            // position P.
            //
            //  0---1
            //  |  /|
            //  | P |
            //  |/  |
            //  2---3
            //
            // 1st triangle: 0, 1, 2
            // 2nd triangle: 2, 1, 3
            //
            // The 4 vertices of the billboard are setup to be located at
            // position P.
            //
            // Each vertex is assigned unit offsets from the billboard's center
            // position (i.e., vertex 0's offset is (1, -1).
            //
            // The billboard vertex shader will transform each billboard's
            // vertex to the correct position in world space.

            BillboardVertex[] vertices = new BillboardVertex[4];

            vertices[0].Position = position;
            vertices[0].TexCoord = new Vector4(0.0f, 0.0f, -1.0f, 1.0f);

            vertices[1].Position = position;
            vertices[1].TexCoord = new Vector4(1.0f, 0.0f, 1.0f, 1.0f);

            vertices[2].Position = position;
            vertices[2].TexCoord = new Vector4(0.0f, 1.0f, -1.0f, -1.0f);

            vertices[3].Position = position;
            vertices[3].TexCoord = new Vector4(1.0f, 1.0f, 1.0f, -1.0f);
                        

            vertexBuffer = new VertexBuffer(graphicsDevice, typeof(BillboardVertex), vertices.Length, BufferUsage.WriteOnly);
            vertexBuffer.SetData(vertices);

            short[] indices =
            {
                (short)0, (short)1, (short)2,
                (short)2, (short)1, (short)3
            };

            indexBuffer = new IndexBuffer(graphicsDevice, IndexElementSize.SixteenBits, indices.Length, BufferUsage.WriteOnly);
            indexBuffer.SetData(indices);
        }

        public void Draw(GraphicsDevice graphicsDevice, Effect effect)
        {
            if (graphicsDevice != null && effect != null)
            {
                graphicsDevice.SetVertexBuffer(vertexBuffer);
                graphicsDevice.Indices = indexBuffer;

                foreach (EffectPass pass in effect.CurrentTechnique.Passes)
                {
                    pass.Apply();
                    graphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, vertexBuffer.VertexCount, 0, 2);
                }

                graphicsDevice.Indices = null;
                graphicsDevice.SetVertexBuffer(null);
            }
        }
    }
}
