//-----------------------------------------------------------------------------
// Copyright (c) 2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Dhpoware;

namespace XNABillboard2
{
    /// <summary>
    /// This XNA Windows application builds on the XNA 4.0 Billboarding Demo
    /// (http://www.dhpoware.com/demos/xnaBillboards.html) by introducing an
    /// alternative shader based billboarding technique. For further details
    /// regarding this technique see "GPU Gems 2: Chapter 1 Toward Photorealism
    /// in Virtual Botany" (http://developer.nvidia.com/node/18). Essentially
    /// this shader based billboarding technique draw the billboard as an
    /// indexed triangle list consisting of 2 separate triangles.
    /// The application setup involves creating a vertex buffer and filling it
    /// with 4 vertices. Each vertex is positioned at the center of the
    /// billboard quad and stores offset describing how far in the x and y
    /// axis the vertex is from the corners of the billboard's quad. Doing this
    /// allows the vertex shader to transform the billboard's vertices in world
    /// space to its correct position in the billboard's quad based on the
    /// camera's x and y axes. This demo also includes the GridComponent class
    /// which is used to render a 3D grid on the world XZ plane.
    /// </summary>
    public class Demo : Microsoft.Xna.Framework.Game
    {
        private static void Main(string[] args)
        {
            using (Demo game = new Demo())
            {
                game.Run();
            }
        }

        private enum BillboardingTechnique
        {
            CameraAligned,
            WorldYAxisAligned
        }

        private CameraComponent cameraComponent;
        private GridComponent gridComponent;
        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;
        private SpriteFont spriteFont;
        private KeyboardState currentKeyboardState;
        private KeyboardState prevKeyboardState;
        private int windowWidth;
        private int windowHeight;
        private int frames;
        private int framesPerSecond;
        private bool displayHelp;
        private TimeSpan elapsedTime = TimeSpan.Zero;
        private Vector2 billboardSize = new Vector2(2.0f, 2.0f);
        private Billboard billboard;
        private Texture2D billboardTexture;
        private Effect billboardEffect;
        private BillboardingTechnique billboardingTechnique;
        
        public Demo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            cameraComponent = new CameraComponent(this);
            Components.Add(cameraComponent);

            gridComponent = new GridComponent(this);
            Components.Add(gridComponent);

            Window.Title = "XNA 4.0 Billboards 2 Demo";
            IsFixedTimeStep = false;
        }

        protected override void Initialize()
        {
            base.Initialize();

            // Setup the window to be a quarter the size of the desktop.
            windowWidth = GraphicsDevice.DisplayMode.Width / 2;
            windowHeight = GraphicsDevice.DisplayMode.Height / 2;

            // Setup frame buffer.
            graphics.SynchronizeWithVerticalRetrace = false;
            graphics.PreferredBackBufferWidth = windowWidth;
            graphics.PreferredBackBufferHeight = windowHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();

            // Setup the initial input states.
            currentKeyboardState = Keyboard.GetState();

            // Setup the GridComponent.
            gridComponent.CreateGrid(Color.LightGray, Vector3.Zero, 16, 1);

            // Setup the camera.
            cameraComponent.CurrentBehavior = Camera.Behavior.Orbit;
            cameraComponent.ClickAndDragMouseRotation = true;
            cameraComponent.Perspective(90.0f, (float)windowWidth / (float)windowHeight, 0.1f, 3.0f * gridComponent.GridSize);
            cameraComponent.LookAt(new Vector3(0.0f, gridComponent.GridSize * 0.25f, gridComponent.GridSize * 0.5f), Vector3.Zero, Vector3.Up);
            cameraComponent.OrbitMinZoom = gridComponent.GridSize * 0.25f;
            cameraComponent.OrbitMaxZoom = gridComponent.GridSize * 1.5f;
            cameraComponent.OrbitOffsetDistance = cameraComponent.Position.Length();

            // Setup the billboard.
            billboard = new Billboard(GraphicsDevice, new Vector3(0.0f, billboardSize.Y * 0.5f, 0.0f), billboardSize.X, billboardSize.Y);
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            
            // Load the sprite font. The sprite font has a 3 pixel outer glow
            // baked into it so we need to decrease the spacing so that the
            // SpriteFont will render correctly.
            spriteFont = Content.Load<SpriteFont>(@"Fonts\DemoFont14pt");
            spriteFont.Spacing = -4.0f;

            // Load texture for the billboard.
            billboardTexture = Content.Load<Texture2D>(@"Textures\grass");

            // Load the Effect file that implements the billboarding shaders.
            billboardEffect = Content.Load<Effect>(@"Effects\Billboard");
            billboardEffect.CurrentTechnique = billboardEffect.Techniques["BillboardingWorldYAxisAligned"];
            billboardingTechnique = BillboardingTechnique.WorldYAxisAligned;
        }

        protected override void UnloadContent()
        {
        }

        private void ToggleFullScreen()
        {
            int newWidth = 0;
            int newHeight = 0;

            graphics.IsFullScreen = !graphics.IsFullScreen;

            if (graphics.IsFullScreen)
            {
                newWidth = GraphicsDevice.DisplayMode.Width;
                newHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                newWidth = windowWidth;
                newHeight = windowHeight;
            }

            graphics.PreferredBackBufferWidth = newWidth;
            graphics.PreferredBackBufferHeight = newHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();
        }

        private bool KeyJustPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && prevKeyboardState.IsKeyUp(key);
        }

        private void ProcessKeyboard()
        {
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            if (KeyJustPressed(Keys.Escape))
                this.Exit();

            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) || currentKeyboardState.IsKeyDown(Keys.RightAlt))
            {
                if (KeyJustPressed(Keys.Enter))
                    ToggleFullScreen();
            }

            if (KeyJustPressed(Keys.H))
                displayHelp = !displayHelp;

            if (KeyJustPressed(Keys.D1))
            {
                billboardingTechnique = BillboardingTechnique.CameraAligned;
                billboardEffect.CurrentTechnique = billboardEffect.Techniques["BillboardingCameraAligned"];
            }

            if (KeyJustPressed(Keys.D2))
            {
                billboardingTechnique = BillboardingTechnique.WorldYAxisAligned;
                billboardEffect.CurrentTechnique = billboardEffect.Techniques["BillboardingWorldYAxisAligned"];
            }
        }

        private void UpdateFrameRate(GameTime gameTime)
        {
            elapsedTime += gameTime.ElapsedGameTime;

            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                framesPerSecond = frames;
                frames = 0;
            }
        }

        private void IncrementFrameCounter()
        {
            ++frames;
        }

        protected override void Update(GameTime gameTime)
        {
            if (!this.IsActive)
                return;

            base.Update(gameTime);

            ProcessKeyboard();
            UpdateFrameRate(gameTime);

            gridComponent.ViewMatrix = cameraComponent.ViewMatrix;
            gridComponent.ProjectionMatrix = cameraComponent.ProjectionMatrix;
        }

        private void DrawText()
        {
            string text = null;

            if (displayHelp)
            {
                System.Text.StringBuilder buffer = new System.Text.StringBuilder();

                buffer.AppendLine("Press 1 for camera-aligned billboarding");
                buffer.AppendLine("Press 2 for world y-axis aligned billboarding");
                buffer.AppendLine("Press ALT + ENTER to toggle full screen");
                buffer.AppendLine("Press H to hide help");

                text = buffer.ToString();
            }
            else
            {
                System.Text.StringBuilder buffer = new System.Text.StringBuilder();

                buffer.AppendFormat("FPS: {0}\n", framesPerSecond);
                buffer.AppendFormat("Camera: x:{0} y:{1} z:{2}\n",
                    cameraComponent.Position.X.ToString("#0.00"),
                    cameraComponent.Position.Y.ToString("#0.00"),
                    cameraComponent.Position.Z.ToString("#0.00"));

                switch (billboardingTechnique)
                {
                case BillboardingTechnique.CameraAligned:
                    buffer.AppendLine("Technique: Camera Aligned");
                    break;

                case BillboardingTechnique.WorldYAxisAligned:
                    buffer.AppendLine("Technique: World Y-Axis Aligned");
                    break;

                default:
                    break;
                }

                buffer.AppendLine();
                buffer.Append("Press H to display help");

                text = buffer.ToString();
            }

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            spriteBatch.DrawString(spriteFont, text, new Vector2(1.0f, 1.0f), Color.Yellow);
            spriteBatch.End();
        }

        private void DrawBillboard()
        {
            RasterizerState prevRasterizerState = GraphicsDevice.RasterizerState;
            BlendState prevBlendState = GraphicsDevice.BlendState;

            GraphicsDevice.BlendState = BlendState.NonPremultiplied;
            GraphicsDevice.RasterizerState = RasterizerState.CullNone;

            billboardEffect.Parameters["world"].SetValue(Matrix.Identity);
            billboardEffect.Parameters["view"].SetValue(cameraComponent.ViewMatrix);
            billboardEffect.Parameters["projection"].SetValue(cameraComponent.ProjectionMatrix);
            billboardEffect.Parameters["billboardSize"].SetValue(billboardSize);
            billboardEffect.Parameters["colorMap"].SetValue(billboardTexture);

            billboard.Draw(GraphicsDevice, billboardEffect);

            GraphicsDevice.BlendState = prevBlendState;
            GraphicsDevice.RasterizerState = prevRasterizerState;
        }

        protected override void Draw(GameTime gameTime)
        {
            if (!this.IsActive)
                return;

            GraphicsDevice.Clear(Color.CornflowerBlue);
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            
            base.Draw(gameTime);

            DrawBillboard();
            DrawText();

            IncrementFrameCounter();
        }
    }
}